/*
 * asmatch.c,v 1.1 1993/12/23 03:11:04 jch Exp
 */

/*%Copyright%*/


#include "include.h"
#ifdef PROTO_INET
#include "inet.h"
#ifdef PROTO_OSPF
#include "ospf.h"
#ifdef	notdef
#include "ospf_rtab.h"
#include "ospf_gated.h"
#endif
#endif /* PROTO_OSPF */
#ifdef PROTO_BGP
#include "bgp.h"
#endif /* PROTO_BGP */
#endif /* PROTO_INET */
#include "parse.h"


/**/
/* AS path matching code */

char *
aspath_adv_origins __PF1(mask, flag_t)
{
    static char line[BUFSIZ];
    
    if (mask == -1) {
	(void) strcpy(line, "all");
    } else {
	bits *p;
	
	*line = (char) 0;
	for (p = path_Orgs; p->t_name; p++) {
	    if ((1 << p->t_bits) & mask) {
		if (*line) {
		    (void) strcat(line, " | ");
		}
		(void) strcat(line, gd_lower(p->t_name));
	    }
	}
    }

    return line;
}

void
aspath_adv_free __PF1(psp, void_t)
{
    task_block_free(aspath_match_index, psp);
}


char *
aspath_adv_print __PF1(psp, void_t)
{
    asmatch_t *asmp = (asmatch_t *) psp;
    static char line[BUFSIZ];
    int i, j;
    asp_table *astp = asmp->nfa;

    sprintf(line, "[ start: ");

    for (i = 0; i < 1 + ASP_ACC / ASP_WORD; i++) {
	sprintf(&line[strlen(line)], "%#x ",
		asmp->start[i]);
    }

    for (i = 0; i < ASP_ACC; i++, astp++) {
	NFA_NONZERO(astp->next, j);
	if (j || astp->as) {
	    sprintf(&line[strlen(line)], "as: %d states: ",
		    astp->as);
	    for (j = 0; j < 1 + ASP_ACC / ASP_WORD; j++) {
		sprintf(&line[strlen(line)], "%#x ",
			astp->next[j]);
	    }
	}
    }
    sprintf(&line[strlen(line)], "] origins %s",
	    aspath_adv_origins(asmp->origin_mask));

    return line;
}


int
aspath_adv_compare __PF2(psp1, void_t,
			 psp2, void_t)
{
    asmatch_t *asmp1 = (asmatch_t *) psp1;
    asmatch_t *asmp2 = (asmatch_t *) psp2;

    /* XXX - need to be a bit more intellegent about this */
    return asmp1 == asmp2;
}


int
aspath_adv_match __PF2(psp, void_t,
		       asp, register as_path *)
{
    register u_short *lp;
    register u_short *p = PATH_SHORT_PTR(asp);
    asmatch_t *asmp = (asmatch_t *) psp;
    flag_t state[1 + ASP_ACC / ASP_WORD], new[1 + ASP_ACC / ASP_WORD];

    if (!asp) {
	return FALSE;
    }

    if (!BIT_TEST(asmp->origin_mask, 1 << asp->path_origin)) {
	return FALSE;
    }

    NFA_ASG(state, asmp->start);

    for (lp = p + (asp->path_len/sizeof (u_short)) - 1;
	 p <= lp;
	 p++)  {
	register int i;
	int nonzero;
	
	NFA_ZERO(new);
	NFA_CLR(state, ASP_ACC);
	for (i = 0; i < ASP_ACC; i++) {
	    if (NFA_ISSET(state, i)) {
		if (!asmp->nfa[i].as || asmp->nfa[i].as == *p) {
		    NFA_OR(new, new, asmp->nfa[i].next);
		    if (p == lp
			&& NFA_ISSET(new, ASP_ACC)) {
			return TRUE;
		    }
		}
		NFA_CLR(state, i);
		NFA_NONZERO(state, nonzero);
		if (!nonzero) {
		    break;
		}
	    }
	}
	NFA_ASG(state, new);
	NFA_NONZERO(state, nonzero);
	if (!nonzero) {
	    break;
	}
    }

    return FALSE;
}


/*
 *  functions used to build an aspath regex
 */

static unsigned	aspath_state;
static asmatch_t *aspath_current;

#define	ASPATH_NEXT_STATE	if (aspath_state++ == ASP_BITS) { \
	return 0; \
    }


void
aspath_init_regex __PF0(void)
{
    aspath_state = 0;
    aspath_current = (asmatch_t *) task_block_alloc(aspath_match_index);
}


asmatch_t *
aspath_consume_current __PF0(void)
{
    asmatch_t *rtn = aspath_current;

    aspath_current = (asmatch_t *) 0;
    return rtn;
}


void
aspath_simple_regex __PF2(regex, asp_stack *,
			new, asp_stack *)
{
    NFA_ASG(regex->start, new->start);
    NFA_ASG(aspath_current->start, regex->start);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	regex : sum %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));
}


void
aspath_copy_regex __PF2(regex, asp_stack *,
			new, asp_stack *)
{
    NFA_ASG(regex->start, new->start);
    NFA_ASG(regex->state, new->state);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	regex : sum %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));
}


void
aspath_merge_regex  __PF3(regex, asp_stack *,
			  new, asp_stack *,
			  old, asp_stack *)
{
    NFA_OR(regex->start, new->start, old->start);
    NFA_OR(regex->state, new->state, old->state);
    NFA_ASG(aspath_current->start, regex->start);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	regex : sum | regex %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));
}


int
aspath_prepend_regex __PF3(regex, asp_stack *,
			    new, asp_stack *,
			    old, asp_stack *)
{
    int i;
  
    NFA_OR(regex->state, new->state, old->state);
  
    for (i = 0; i < ASP_ACC; i++) {
	int nonzero;
    
	if (NFA_ISSET(new->state, i)) {
	    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
		NFA_CLR(aspath_current->nfa[i].next, ASP_ACC);
		NFA_OR(aspath_current->nfa[i].next,
		       aspath_current->nfa[i].next, old->start);
	    }
	    NFA_CLR(new->state, i);
	    NFA_NONZERO(new->state, nonzero);
	    if (nonzero) {
		break;
	    }
	}
    }
    if (i > ASP_ACC) {
	return 0;
    }
    if (NFA_ISSET(new->start, ASP_ACC)) {
	NFA_CLR(new->start, ASP_ACC);
	NFA_OR(new->start, new->start, old->start);
    }
    NFA_ASG(regex->start, new->start);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	sum : term sum %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));

  return 1;
}

void
aspath_zero_or_more_term __PF2(regex, asp_stack *,
			       old, asp_stack *)
{
    int i;
  
    for (i = 0; i < ASP_ACC; i++) {
	if (NFA_ISSET(old->state, i)) {
	    int nonzero;
      
	    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
		NFA_OR(aspath_current->nfa[i].next,
		       aspath_current->nfa[i].next, old->start);
	    }
	    NFA_CLR(old->state, i);
	    NFA_NONZERO(old->state, nonzero);
	    if (nonzero) {
		break;
	    }
	}
    }
    NFA_ASG(regex->start, old->start);
    NFA_SET(regex->start, ASP_ACC);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	term : symbol * %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));
}


void
aspath_one_or_more_term __PF2(regex, asp_stack *,
			      old, asp_stack *)
{
    int i;
  
    for (i = 0; i < ASP_ACC; i++) {
	if (NFA_ISSET(old->state, i)) {
	    int nonzero;
      
	    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
		NFA_OR(aspath_current->nfa[i].next,
		       aspath_current->nfa[i].next, old->start);
	    }
	    NFA_CLR(old->state, i);
	    NFA_NONZERO(old->state, nonzero);
	    if (nonzero) {
		break;
	    }
	}
    }
    NFA_ASG(regex->start, old->start);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	term : symbol + %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));
}


void
aspath_zero_or_one_term __PF2(regex, asp_stack *,
                              old, asp_stack *)
{
    NFA_ASG(regex->start, old->start);
    NFA_SET(regex->start, ASP_ACC);
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	term : symbol ? %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));
}


int
aspath_range_term __PF3(regex, asp_stack *,
			old, asp_stack *,
			range, asp_range *)
{
    u_int copy;
    flag_t finalstate[1 + ASP_ACC / ASP_WORD];
    flag_t newstate[1 + ASP_ACC / ASP_WORD];
    flag_t oldstate[1 + ASP_ACC / ASP_WORD];
    flag_t newstart[1 + ASP_ACC / ASP_WORD];
  
    NFA_ASG(finalstate, old->state);
    NFA_ASG(newstate, old->state);
    NFA_ASG(regex->start, old->start);
    NFA_ASG(newstart, old->start);
    if (!range->begin) {
	NFA_SET(regex->start, ASP_ACC);
    }
    for (copy = 1; copy < range->begin; copy++) {
	int i;
    
	NFA_ASG(oldstate, old->state);
	NFA_ZERO(newstart);
	NFA_ZERO(newstate);
	for (i = 0; i < ASP_ACC; i++) {
	    if (NFA_ISSET(old->state, i)) {
		aspath_current->nfa[aspath_state].as = aspath_current->nfa[i].as;
		NFA_ASG(aspath_current->nfa[aspath_state].next,
			aspath_current->nfa[i].next);
		NFA_SHIFT(aspath_current->nfa[aspath_state].next, aspath_state-i);
		NFA_SET(newstate, aspath_state);
		if (NFA_ISSET(old->start, i)) {
		    int nonzero;
	  
		    NFA_SET(newstart, aspath_state);
		    ASPATH_NEXT_STATE;
		    NFA_CLR(old->state, i);
		    NFA_NONZERO(old->state, nonzero);
		    if (nonzero) {
			break;
		    }
		}
	    }
	    if (NFA_ISSET(regex->start, ASP_ACC)) {
		NFA_OR(regex->start,regex->start,newstart);
	    }
	    for (i = 0; i < ASP_ACC; i++) {
		if (NFA_ISSET(oldstate, i)) {
		    int nonzero;
	  
		    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
			if (copy < range->begin) {
			    NFA_CLR(aspath_current->nfa[i].next, ASP_ACC);
			}
			NFA_OR(aspath_current->nfa[i].next,
			       aspath_current->nfa[i].next, newstart);
		    }
		    NFA_CLR(oldstate, i);
		    NFA_NONZERO(oldstate, nonzero);
		    if (nonzero) {
			break;
		    }
		}
	    }
	    NFA_OR(finalstate, finalstate, newstate);
	    NFA_ASG(old->start, newstart);
	    NFA_ASG(old->state, newstate);
	}
	if (range->end) {
	    for (i = 0; i < ASP_ACC; i++) {
		if (NFA_ISSET(newstate, i)) {
		    int nonzero;
	  
		    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
			NFA_OR(aspath_current->nfa[i].next,
			       aspath_current->nfa[i].next, newstart);
		    }
		    NFA_CLR(newstate, i);
		    NFA_NONZERO(newstate, nonzero);
		    if (nonzero) {
			break;
		    }
		}
	    }
	}	
	NFA_ASG(regex->state, finalstate);
	trace_tf(trace_global,
		 TR_PARSE,
		 0,
		 ("%s: REDUCED:	term : symbol { range } %s",
		  parse_where(),
		  aspath_adv_print(aspath_current)));
    }

    return 1;
}


int
aspath_as_transition __PF2(regex, asp_stack *,
			   as_num, int)
{
    NFA_ZERO(regex->start);
    NFA_ZERO(regex->state);
    NFA_SET(regex->start, aspath_state);
    NFA_SET(regex->state, aspath_state);
    aspath_current->nfa[aspath_state].as = htons(as_num);
    NFA_SET(aspath_current->nfa[aspath_state].next, ASP_ACC);
    ASPATH_NEXT_STATE;
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	symbol : %d %s",
	      parse_where(),
	      as_num,
	      aspath_adv_print(aspath_current)));

  return 1;
}


int
aspath_any_transition __PF1(regex, asp_stack *)
{
    NFA_ZERO(regex->start);
    NFA_ZERO(regex->state);
    NFA_SET(regex->start, aspath_state);
    NFA_SET(regex->state, aspath_state);
    NFA_SET(aspath_current->nfa[aspath_state].next, ASP_ACC);
    aspath_current->nfa[aspath_state].as = 0;
    ASPATH_NEXT_STATE;
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("%s: REDUCED:	symbol : . %s",
	      parse_where(),
	      aspath_adv_print(aspath_current)));

    return 1;
}


/*
 * ------------------------------------------------------------------------
 * 
 * 	GateD, Release 3
 * 
 * 	Copyright (c) 1990,1991,1992,1993 by Cornell University
 * 	    All rights reserved.
 * 
 * 	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
 * 	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * 	LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * 	AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 	Royalty-free licenses to redistribute GateD Release
 * 	3 in whole or in part may be obtained by writing to:
 * 
 * 	    GateDaemon Project
 * 	    Information Technologies/Network Resources
 * 	    200 CCC
 * 	    Cornell University
 * 	    Ithaca, NY  14853-2601  USA
 * 
 * 	GateD is based on Kirton's EGP, UC Berkeley's routing
 * 	daemon	 (routed), and DCN's HELLO routing Protocol.
 * 	Development of GateD has been supported in part by the
 * 	National Science Foundation.
 * 
 * 	Please forward bug fixes, enhancements and questions to the
 * 	gated mailing list: gated-people@gated.cornell.edu.
 * 
 * 	Authors:
 * 
 * 		Jeffrey C Honig <jch@gated.cornell.edu>
 * 		Scott W Brim <swb@gated.cornell.edu>
 * 
 * ------------------------------------------------------------------------
 * 
 *       Portions of this software may fall under the following
 *       copyrights:
 * 
 * 	Copyright (c) 1988 Regents of the University of California.
 * 	All rights reserved.
 * 
 * 	Redistribution and use in source and binary forms are
 * 	permitted provided that the above copyright notice and
 * 	this paragraph are duplicated in all such forms and that
 * 	any documentation, advertising materials, and other
 * 	materials related to such distribution and use
 * 	acknowledge that the software was developed by the
 * 	University of California, Berkeley.  The name of the
 * 	University may not be used to endorse or promote
 * 	products derived from this software without specific
 * 	prior written permission.  THIS SOFTWARE IS PROVIDED
 * 	``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * 	INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * 	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
